use std::fmt;

pub enum Piece {
    LightPawn,
    LightKing,
    DarkPawn,
    DarkKing,
}

pub enum BoardTile {
    Empty,
    Piece(Piece),
}

pub enum BoardSize {
    American,
    International,
    Canadian,
}

impl BoardSize {
    fn get_size(&self) -> usize {
        match &self {
            BoardSize::American => 8,
            BoardSize::International => 10,
            BoardSize::Canadian => 12,
        }
    }
}

pub struct Board {
    dimensions: BoardSize,
    tiles: Vec<BoardTile>,
}

impl Board {
    pub fn new(size: BoardSize) -> Board {
        let board_size = size.get_size();
        let width = board_size / 2;
        let height = board_size;
        let starting_rows = board_size / 2 - 1;

        let playing_tiles = (0..width * height)
            .map(|tile| {
                let row = tile / width;
                match row {
                    r if r < starting_rows => BoardTile::Piece(Piece::LightPawn),
                    r if r > height - starting_rows - 1 => BoardTile::Piece(Piece::DarkPawn),
                    _ => BoardTile::Empty,
                }
            })
            .collect();

        Board {
            dimensions: size,
            tiles: playing_tiles,
        }
    }

    fn get_index(&self, row: usize, column: usize) -> usize {
        let width = self.dimensions.get_size();
        (row * (width / 2) + column) as usize
    }
}

static COLUMN_COORDINATE_BORDER: &str = " ABCDEFGHIJKL";

impl fmt::Display for Board {
    fn fmt(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
        let width = self.dimensions.get_size();

        // Create the top level border.
        let (column_border_chars, _) = COLUMN_COORDINATE_BORDER.split_at(width + 1);
        writeln!(formatter, "{}", column_border_chars)?;

        for row in 0..width {
            write!(formatter, "{}", (row + 1).to_string())?;

            for col in 0..(width / 2) {
                let idx = self.get_index(row, col);
                let token = match &self.tiles[idx] {
                    BoardTile::Piece(x) => match x {
                        Piece::LightPawn => "○",
                        Piece::LightKing => "⍟",
                        Piece::DarkPawn => "●",
                        Piece::DarkKing => "✪",
                    },
                    BoardTile::Empty => "◰",
                };

                if row % 2 == 0 {
                    write!(formatter, " {}", token)?;
                } else {
                    write!(formatter, "{} ", token)?;
                }
            }

            write!(formatter, "\n")?;
        }

        Ok(())
    }
}

pub enum GameState {
    LightPlayerTurn,
    DarkPlayerTurn,
    LightPlayerWin,
    DarkPlayerWin,
    DrawGame,
}

pub struct Game {
    board: Board,
    state: GameState,
}

impl Game {
    pub fn new(size: BoardSize) -> Game {
        // TODO: Add the concept of rulesets for the various checkers variants.
        Game {
            board: Board::new(size),
            state: GameState::LightPlayerTurn,
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::Board;
    use crate::BoardSize;

    #[test]
    fn initial_international_board_setup() {
        let expected = " \
                        ABCDEFGHIJ\n\
                        1 ○ ○ ○ ○ ○\n\
                        2○ ○ ○ ○ ○ \n\
                        3 ○ ○ ○ ○ ○\n\
                        4○ ○ ○ ○ ○ \n\
                        5 ◰ ◰ ◰ ◰ ◰\n\
                        6◰ ◰ ◰ ◰ ◰ \n\
                        7 ● ● ● ● ●\n\
                        8● ● ● ● ● \n\
                        9 ● ● ● ● ●\n\
                        10● ● ● ● ● \n\
                        ";
        assert_eq!(Board::new(BoardSize::International).to_string(), expected);
    }

    #[test]
    fn initial_american_board_setup() {
        let expected = " \
                        ABCDEFGH\n\
                        1 ○ ○ ○ ○\n\
                        2○ ○ ○ ○ \n\
                        3 ○ ○ ○ ○\n\
                        4◰ ◰ ◰ ◰ \n\
                        5 ◰ ◰ ◰ ◰\n\
                        6● ● ● ● \n\
                        7 ● ● ● ●\n\
                        8● ● ● ● \n\
                        ";
        assert_eq!(Board::new(BoardSize::American).to_string(), expected);
    }

    #[test]
    fn initial_canadian_board_setup() {
        let expected = " \
                        ABCDEFGHIJKL\n\
                        1 ○ ○ ○ ○ ○ ○\n\
                        2○ ○ ○ ○ ○ ○ \n\
                        3 ○ ○ ○ ○ ○ ○\n\
                        4○ ○ ○ ○ ○ ○ \n\
                        5 ○ ○ ○ ○ ○ ○\n\
                        6◰ ◰ ◰ ◰ ◰ ◰ \n\
                        7 ◰ ◰ ◰ ◰ ◰ ◰\n\
                        8● ● ● ● ● ● \n\
                        9 ● ● ● ● ● ●\n\
                        10● ● ● ● ● ● \n\
                        11 ● ● ● ● ● ●\n\
                        12● ● ● ● ● ● \n\
                        ";
        assert_eq!(Board::new(BoardSize::Canadian).to_string(), expected);
    }
}
